import getopt, sys,math,time, ctypes
from nifti import NiftiImage
import CVtypes as cv
import scipy as sc
from ctypes import sizeof, cast




def cc_process(output, input, verbose):
    cvimg = nifti2opencv(input, verbose)
    orig_img = cv.cvCloneImage(cvimg)
    #segmentation_and_contour(cvimg, True)
    #segmentation(cvimg, True)
    #threshold(cvimg, True)
    cvimg, contour, area = segmentation_and_contour(cvimg, verbose)
    if verbose:
        cv.cvReleaseImage(cvimg)
        cvimg = cv.cvCloneImage(orig_img)
        curvature_small = getCurvature(contour, 5, cvimg)
        cv.cvReleaseImage(cvimg)
        cvimg = cv.cvCloneImage(orig_img)
        curvature_large = getCurvature(contour, 20, cvimg)
        cv.cvReleaseImage(cvimg)
        cvimg = cv.cvCloneImage(orig_img)
        ends = findEnds(curvature_small, curvature_large, contour, cv.cvGetSize(cvimg), cvimg)
        cv.cvReleaseImage(cvimg)
        cvimg = cv.cvCloneImage(orig_img)
        axis = findAxisCenter(contour, ends, cvimg)
    else:
        curvature_small = getCurvature(contour, 5)
        curvature_large = getCurvature(contour, 10)
        ends = findEnds(curvature_small, curvature_large, contour, cv.cvGetSize(cvimg))
        axis = findAxisCenter(contour, ends)
    cv.cvReleaseImage(cvimg)
    cvimg = cv.cvCloneImage(orig_img)
    selected_points, cvimg = divideAxis(axis, 15, cvimg, verbose)
    angles = calcAngles(selected_points)
    printOutput(selected_points, angles, area, input)
    if output != None:
        cv.cvSaveImage(output, cvimg)
    cv.cvReleaseImage(cvimg)
    cv.cvReleaseImage(orig_img)

def usage():
    print(sys.argv[0] + " --verbose --help --output out.nii.gz in.nii.gz")

def main():
    try:
        opts, args = getopt.getopt(sys.argv[1:], "ho:v", ["help", "output=","verbose"])
    except getopt.GetoptError, err:
        # print help information and exit:
        print str(err) # will print something like "option -a not recognized"
        usage()
        return 2
        
    if(len(args) != 1):
        usage()
        return 2
    else:
        input = args[0]
        
    output = None
    verbose = False
    for o, a in opts:
        if o in ("-v", "--verbose"):
            verbose = True
        elif o in ("-h", "--help"):
            usage()
            sys.exit()
        elif o in ("-o", "--output"):
            output = a
        else:
            assert False, "unhandled option"
    
    cc_process(output, input, verbose)
    
def slice2opencv(slice, depth):
    cvimg = cv.cvCreateImage(cv.CvSize(slice.shape[1], slice.shape[0]), depth, 1)
    for x in range(len(slice)):
        for y in range(len(slice[x])):
            cv.cvSet2D(cvimg, x, y, cv.CvScalar(float(slice[x][y])))
    return cvimg

def nifti2opencv(nifti_filename, verbose=False):
    middle_slice = 126
    nim3d = NiftiImage(nifti_filename)
    if nim3d.header['datatype'] == 1:
        #binary
        depth = cv.IPL_DEPTH_8U
    elif nim3d.header['datatype'] == 2:
        depth = cv.IPL_DEPTH_8U
    elif nim3d.header['datatype'] == 4:
        depth = cv.IPL_DEPTH_16S
    elif nim3d.header['datatype'] == 8:
        depth = cv.IPL_DEPTH_32S
    elif nim3d.header['datatype'] == 16:
        depth = cv.IPL_DEPTH_32F
    elif nim3d.header['datatype'] == 64:
        depth = cv.IPL_DEPTH_64F
    else:
        assert False, "unknown datatype: " + nim3d.header['datatype']
     
    slice = nim3d.data[::-1, :, middle_slice]
    
    cvimg = slice2opencv(slice, depth)
        
    if verbose:     
        showImg("original", cvimg)
        
    return cvimg
    
def segmentation(cvimg, verbose = False):

    max_value = 0;
    index_of_max_value = 0;
    for i in xrange((cv.cvGetSize(cvimg)).height):
        value = (cv.cvGet2D(cvimg,i,170)).val[0]
        if value >= max_value:
            max_value = value
            index_of_max_value = i
    #print index_of_max_value
    seedPoint = cv.CvPoint(170,index_of_max_value)
    cv.cvFloodFill(cvimg, seedPoint, cv.CvScalar(255), cv.CvScalar(10), cv.CvScalar(11))
    cv.cvSet2D(cvimg, seedPoint.y, seedPoint.x, cv.CvScalar(255))
    if verbose:        
        showImg("segmented", cvimg)
    return cvimg

def segmentation_and_contour(cvimg, verbose = False):
    max_value = 0;
    index_of_max_value = 0;
    queried_column = 165
    th_range = (1,50)
    
    #finding seedPoint
    for i in xrange((cv.cvGetSize(cvimg)).height):
        value = (cv.cvGet2D(cvimg,i,queried_column)).val[0]
        if value >= max_value:
            max_value = value
            index_of_max_value = i
    seedPoint = cv.CvPoint(queried_column,index_of_max_value)
    
    best_area = None;
    prev_area = None;
    prev_contours_ptr = None;
    prev_max_right = None;
    th_cvimg = cv.cvCloneImage(cvimg)
    for low_diff in range(th_range[0],th_range[1]):
        cv.cvCopy(cvimg, th_cvimg)
        cv.cvFloodFill(th_cvimg, seedPoint, cv.CvScalar(255), cv.CvScalar(low_diff), cv.CvScalar(low_diff),None,8+cv.CV_FLOODFILL_FIXED_RANGE)
        area = valueCount(255, th_cvimg)
        
        if area >= 350:
            storage = cv.cvCreateMemStorage(0)
            contours_ptr = cv.POINTER(cv.CvSeq)()
            cv.cvThreshold(th_cvimg, th_cvimg, 254, 255, cv.CV_THRESH_BINARY)
            num_contours = cv.cvFindContours(th_cvimg, storage, cv.byref(contours_ptr), sizeof(cv.CvContour), cv.CV_RETR_EXTERNAL, cv.CV_CHAIN_APPROX_NONE, cv.CvPoint(0,0)) #@UnusedVariable
            if num_contours == 1:
                print area
                best_area = prev_area
                best_contours_ptr = prev_contours_ptr
                
                max_right = 0
                for i in range(contours_ptr[0].total):
                    f = cv.cvGetSeqElem(contours_ptr[0], i)
                    pt = cast(f, ctypes.POINTER(cv.CvPoint))
                    if pt.contents.x > max_right:
                        max_right = pt.contents.x
                print max_right
                if(prev_max_right != None and (max_right-prev_max_right) >=2 and area < 450):
                    break;
                if(prev_area != None and (area-prev_area) > 50 and prev_area < 400):  
                    break;
                if(prev_area != None and (area-prev_area) > 40 and prev_area < 450):  
                    break;
                if(prev_area != None and (area-prev_area) > 17 and prev_area > 450):  
                    break;
                
                else:
                    prev_area = area;
                    prev_contours_ptr = contours_ptr
                    prev_max_right = max_right
                    
        if best_area == None:
            best_area = prev_area
            best_contours_ptr = prev_contours_ptr
    if verbose:
        cv.cvDrawContours(cvimg, best_contours_ptr, cv.CvScalar(255), cv.CvScalar(255), 3,1,8, cv.CvPoint(0,0))     
        showImg("contours" +str(low_diff), cvimg)
        
    contour_list = list()
    for i in range(best_contours_ptr[0].total):
        f = cv.cvGetSeqElem(best_contours_ptr[0], i)
        pt = cast(f, ctypes.POINTER(cv.CvPoint))
        contour_list.append((pt.contents.x,pt.contents.y))
                
    cv.cvReleaseImage(th_cvimg)
    return (cvimg, contour_list, best_area)

def valueCount(value, cvimg):
    counter = 0;
    for y in range(cv.cvGetSize(cvimg).height):
        for x in range(cv.cvGetSize(cvimg).width):
            if cv.cvGet2D(cvimg, y,x).val[0] == value:
                counter += 1
    return counter
        
def threshold(cvimg, verbose = False):
    cv.cvThreshold(cvimg, cvimg, 254, 255, cv.CV_THRESH_BINARY)
    if verbose:        
        showImg("threshold", cvimg)
    return cvimg
def contours(cvimg, verbose = False):
    storage = cv.cvCreateMemStorage(0)
    contours_ptr = cv.POINTER(cv.CvSeq)()
    num_contours = cv.cvFindContours(cvimg, storage, cv.byref(contours_ptr), sizeof(cv.CvContour), cv.CV_RETR_LIST, cv.CV_CHAIN_APPROX_NONE, cv.CvPoint(0,0)) #@UnusedVariable
    longest_contour = 0
    longest_contour_index = 0
    for i in xrange(num_contours):
        #print contours_ptr[i].total
        if  contours_ptr[i].total >= longest_contour:
            longest_contour = contours_ptr[i].total
            longest_contour_index = i
    contour_list = list()
    area = abs(cv.cvContourArea(contours_ptr, slice=cv.CV_WHOLE_SEQ))
    #print longest_contour 
    for i in range(contours_ptr[longest_contour_index].total):
        f = cv.cvGetSeqElem(contours_ptr[longest_contour_index], i)
        pt = cast(f, ctypes.POINTER(cv.CvPoint))
        contour_list.append((pt.contents.x,pt.contents.y))
    cv.cvZero(cvimg)
    cv.cvDrawContours(cvimg, contours_ptr, cv.CvScalar(255), cv.CvScalar(255), 3,1,8, cv.CvPoint(0,0))
    cv.cvNot(cvimg, cvimg)
    if verbose:        
        showImg("contours", cvimg)
    cv.cvReleaseMemStorage(storage)
            
    return (cvimg, contour_list, area)

def distanceMap(cvimg, verbose = False):
    dst = cv.cvCreateImage(cv.cvGetSize(cvimg), cv.IPL_DEPTH_32F, 1)
    bin = cv.cvCreateImage(cv.cvGetSize(cvimg), cv.IPL_DEPTH_8U, 1)
    cv.cvThreshold(cvimg, bin, 1, 1, cv.CV_THRESH_BINARY)
    cv.cvDistTransform(bin, dst, 2 ,3, None, None)
    if verbose:   
        scaled = cv.cvClone(dst)     
        cv.cvConvertScale(scaled, scaled, 0.15,0)
        showImg("distance map", scaled)
    return dst

def getCurvature(contour, k, cvimg=None):
    curvature_list = list()
    for i in range(len(contour)):
        x = contour[i][0]
        x_pred = contour[i-k][0]
        x_succ = contour[(i+k) % len(contour)][0]
        a_2, a_1, a_0 = sc.polyfit([1, 0, 2], [x_pred, x, x_succ], 2) #@UnusedVariable
        
        y = contour[i][1]
        y_pred = contour[i-k][1]
        y_succ = contour[(i+k) % len(contour)][1]
        b_2, b_1, b_0 = sc.polyfit([1, 0, 2], [y_pred, y, y_succ], 2) #@UnusedVariable
        
        curv = 2*(a_1*b_2 - b_1*a_2)/math.pow(a_1**2 +b_1**2, 1.5)
        #print "x = %d, y = %d, curvature = %f" % (x,y,curv)

        curvature_list.append(curv)
    
    if cvimg != None:
        cv.cvZero(cvimg)
        for i in range(len(curvature_list)):
            curv = curvature_list[i]
            x = contour[i][0]
            y = contour[i][1]
            curv_norm = (curv - min(curvature_list))/(max(curvature_list)-min(curvature_list))*255
            cv.cvSet2D(cvimg, y, x, cv.CvScalar(curv_norm))
        showImg("curvature", cvimg)
    return curvature_list

def findEnds(curvature_small, curvature_large, contour, img_size, cvimg=None):
    max_left = 0
    max_left_coo = None
    max_right = 0
    max_right_coo = None
    
    
    left_end = 65000
    right_end = 0
    for i in range(len(curvature_small)): 
        x = contour[i][0]
        if x < left_end:
            left_end = x
        if x > right_end:
            right_end = x
        
    for i in range(len(curvature_small)): 
        x = contour[i][0]
        y = contour[i][1]
        if x < left_end + (right_end-left_end)/7:
            if max_left < curvature_large[i]:
                max_left = curvature_large[i]
                max_left_coo = (x, y)
        elif x > right_end - (right_end-left_end)/3:
            if max_right < curvature_small[i]:
                max_right = curvature_small[i]
                max_right_coo = (x, y)
    if cvimg != None:
        cv.cvSet2D(cvimg, max_left_coo[1], max_left_coo[0], cv.CvScalar(255))
        cv.cvSet2D(cvimg, max_right_coo[1], max_right_coo[0], cv.CvScalar(255))
        showImg("ends", cvimg)
    
    return (max_left_coo, max_right_coo)

def findAxisCenter(contour, ends, cvimg=None):
    left_end_index = contour.index(ends[0])
    right_end_index = contour.index(ends[1])
    lower_contour = list()
    upper_contour = list()
    # TODO
    if left_end_index < right_end_index:
        lower_contour = contour[left_end_index+1:right_end_index-1]
        upper_contour = contour[right_end_index+1:] + contour[:left_end_index-1]
    else:
        lower_contour = contour[left_end_index+1:] + contour[:right_end_index-1]
        upper_contour = contour[right_end_index+1:left_end_index-1]
        
    last_index = 1
    axis = list()
    axis.append(ends[1])
    for i in range(len(upper_contour)):
        min_dist = 1000000
        min_dist_vec = 0
        index = 0

        for j in range(len(lower_contour)):
            vec = (upper_contour[i][0]-lower_contour[j][0],  upper_contour[i][1]-lower_contour[j][1])
            dist = math.sqrt(vec[0]**2 + vec[1]**2)
            if(dist < min_dist):
                min_dist = dist
                min_dist_vec = vec
                index = j
        last_index = index
        axis_point = (upper_contour[i][0]-min_dist_vec[0]/2.0, upper_contour[i][1]-min_dist_vec[1]/2.0,min_dist)
        axis.append(axis_point)
    axis.append(ends[0])
    if cvimg != None:
        for i in range(len(axis)):
            cv.cvSet2D(cvimg, int(round(axis[i][1],0)), int(round(axis[i][0],0)), cv.CvScalar(0))
        for i in range(len(lower_contour)):
            cv.cvSet2D(cvimg, int(lower_contour[i][1]), int(lower_contour[i][0]), cv.CvScalar(100))
        for i in range(len(upper_contour)):
            cv.cvSet2D(cvimg, int(upper_contour[i][1]), int(upper_contour[i][0]), cv.CvScalar(250))
        showImg("axis", cvimg)
    return axis

def divideAxis(axis,n_parts, cvimg, verbose=False):
    total_len = 0
    for i in range(1,len(axis)):
        loc_len = math.sqrt(math.pow(axis[i][0]-axis[i-1][0], 2)+math.pow(axis[i][1]-axis[i-1][1], 2))
        total_len += loc_len
    
    delta = total_len/n_parts
    #print delta
    selected_points = list()
    selected_points.append(axis[0])
    last_index = 0
    for i in range(len(axis)):
        loc_len = math.sqrt(math.pow(axis[i][0]-axis[last_index][0], 2)+math.pow(axis[i][1]-axis[last_index][1], 2))
        if loc_len >= delta:
            last_index = i
            selected_points.append(axis[i])
            
    selected_points.pop()
    selected_points.append(axis[len(axis)-1])
    
    
    for i in range(len(selected_points)-1):
        pt1 = cv.CvPoint(int(round(selected_points[i][0],0)), int(round(selected_points[i][1],0)))
        pt2 = cv.CvPoint(int(round(selected_points[i+1][0],0)), int(round(selected_points[i+1][1],0)))
        cv.cvLine(cvimg, pt1,pt2, cv.CvScalar(0))
        cv.cvSet2D(cvimg, int(round(selected_points[i][1],0)), int(round(selected_points[i][0],0)), cv.CvScalar(255))
    cv.cvSet2D(cvimg, int(round(selected_points[len(selected_points)-1][1],0)), int(round(selected_points[len(selected_points)-1][0],0)), cv.CvScalar(255))
    if verbose:
        showImg("selected points", cvimg)
    return selected_points, cvimg

def calcAngles(segmentsPoints):
    angles = list()
    for i in range(1,len(segmentsPoints)-1):
        vec1 = (segmentsPoints[i][0]-segmentsPoints[i-1][0], segmentsPoints[i][1]-segmentsPoints[i-1][1])
        vec2 = (segmentsPoints[i][0]-segmentsPoints[i+1][0], segmentsPoints[i][1]-segmentsPoints[i+1][1])
        angles.append(math.degrees(segmentsAngle(vec1[0], vec1[1], vec2[0], vec2[1])))
    return angles
def segmentsAngle(a1,a2,b1,b2):
    a_len = math.sqrt(a1**2 + a2**2)
    b_len = math.sqrt(b1**2 + b2**2)
    angle = float(a1*b1+a2*b2)/float(a_len*b_len)
    if angle < -1:
        angle = -1
    if angle > 1:
        angle = 1
    print math.degrees(math.acos(angle))
    return math.acos(angle)

def printOutput(selected_points, angles,area, input_filename):
    print "%s, " %(input_filename),
    print "%f, " %(area),
    print "%f, %f, " %(selected_points[0][0], selected_points[0][1]),
    for i in range(1,len(selected_points)-1):
        print "%f, %f, %f, %f, " % (selected_points[i][0], selected_points[i][1], selected_points[i][2],  angles[i-1]),
    print "%f, %f" %(selected_points[len(selected_points)-1][0], selected_points[len(selected_points)-1][1]),
    

def findAxisSnake(contour, ends, cvimg,orig, axis):
    contour_points = (cv.CvPoint * len(contour))()
    for i in range(len(axis)):
        pt = axis[i]
        contour_points[i] = cv.CvPoint(int(round(pt[0],0)),int(round(pt[1],0)))
        print contour_points[i]
    bin = cv.cvCreateImage(cv.cvGetSize(cvimg), cv.IPL_DEPTH_8U, 1)
    cv.cvThreshold(cvimg, bin, 1, 1, cv.CV_THRESH_BINARY)
    #TODO
    #cvSnakeImage(dist_map, contour_points[0], len(contour_points), x,x,x, CV_VALUE, cvSize(3,3), cvTermCriteria(CV_TERMCRIT_ITER, 100,0.0))
    cv.cvSnakeImage(bin, contour_points, len(contour_points), cv.c_float(0.8),cv.c_float(0.5),cv.c_float(-1), cv.CV_VALUE, cv.CvSize(3,3), cv.CvTermCriteria(cv.CV_TERMCRIT_ITER, 100 ,0.0),1)
    for i in range(len(contour)):
        cv.cvSet2D(orig, contour_points[i].y, contour_points[i].x, cv.CvScalar(250))
        print contour_points[i]
    showImg("snake", orig)
        
def showImg(label, image, period = 0):
    cv.cvNamedWindow(label)
    cv.cvShowImage(label, image)
    if period != 0:
        time.sleep(period)
    else:
        cv.cvWaitKey()

if __name__ == '__main__':
    sys.exit(main())
